open Instructions

type valeurVm =
  | VmConst of int
  | VmOp of string
  | VmClosure of string * (Instructions.instruction list) * (string * valeurVm) list
  | VmContinue of (Instructions.instruction list) * (string * valeurVm) list



let rec valueVm_to_string = function
  | VmClosure (x,cod,env) ->
(*    let env'=List.map (function s,v -> s^"=>"^(valueVm_to_string v)^",") env in *)
(*    let cod'=List.map instr_to_string cod in *)
(*      "Closure("^x^","^(Instructions.to_string cod)^","^(List.fold_left (^) "" env')^")" *)
     "Closure("^x^","^(Instructions.to_string cod)^","^(to_string_env env)^")"
  | VmConst c -> string_of_int c
  | VmOp  s -> s
  | VmContinue(cod,env) -> "("^(Instructions.to_string cod)^","^(to_string_env env)^")"
 
 and 
   to_string_env env =
    let env'=List.map (function s,v -> s^"=>"^(valueVm_to_string v)^",") env in
     "{"^(List.fold_left (^) "" env')^"}"
 and 
   to_string_stack stack =
    let stack'=List.map (function v -> (valueVm_to_string v)^",") stack in
     "[["^(List.fold_left (^) "" stack')^"]]"
  
  
let print_vm code env stack =
 begin
   print_string ("CODE="^(Instructions.to_string code));
   print_newline(); 
   print_string ("ENV="^(to_string_env env));
   print_newline(); 
   print_string ("STACK="^(to_string_stack stack));
   print_newline(); 
   print_newline()
 end

let rec execute code env stack = 
 print_vm code env stack;
 match code with
    [] -> stack
  | i::next -> 
    begin
     match i with
      IVar x -> execute next env ((List.assoc x env)::stack)
     | IConst n -> execute next env ((VmConst n)::stack)
     | IOp op -> execute next env ((VmOp op)::stack)
     | IClos (s, code') -> execute next env (VmClosure(s,code',env)::stack)
     | ILet s -> begin 
                  match stack with
                    hd::tl -> execute next ((s,hd)::env) tl
                 end
     | IEndLet s -> execute next (List.remove_assoc s env) stack
     | IRet -> begin
                match stack with
                 v::VmContinue(c,env')::tl -> execute c env' (v::tl)
               end 
     | IApp -> begin
                match stack with
                   v::VmOp(s)::tl -> 
                    begin
                     match s with
                       "incr" -> let VmConst n =v in 
                                  execute next env (VmConst(n+1)::tl)
                      | _ -> failwith "Operateur inconnu"
                    end
                 | v::VmClosure(s,c,env')::tl -> execute c ((s,v)::env') (VmContinue(next,env)::tl)
               end

   end 
  

let _ =
 let lexbuf = Lexing.from_channel stdin in
 let code = ParserInstr.program LexerInstr.token lexbuf in
 print_newline(); 
 print_string (Instructions.to_string code);
 print_newline(); 
 print_newline();   
  let [v] = execute code [] [] in
    print_string (valueVm_to_string v);
    print_newline(); 
    flush stdout
